"""
The package of the partially-observed time-series imputation model ETSformer.

Refer to the paper
`Gerald Woo, Chenghao Liu, Doyen Sahoo, Akshat Kumar, and Steven Hoi.
ETSformer: Exponential smoothing transformers for time-series forecasting.
In ICLR, 2023.
<https://openreview.net/pdf?id=5m_3whfo483>`_

"""

# Created by Wenjie Du <wenjay.du@gmail.com>
# License: BSD-3-Clause


from .model import ETSformer

__all__ = [
    "ETSformer",
]
