"""
The package of the partially-observed time-series imputation method Median.
"""

# Created by Wenjie Du <wenjay.du@gmail.com>
# License: BSD-3-Clause

from .model import Median

__all__ = [
    "Median",
]
